import inspect
import json
import logging
import re
from collections.abc import Iterable, Mapping

import scrapy
from scrapy import signals

logger = logging.getLogger(__name__)

DEFAULT_REGEXES = [
    r"(?i).*(api[\W_]*key).*",  # apikey and variations e.g: shub_apikey or SC_APIKEY
    r"(?i).*(AWS[\W_]*(SECRET[\W_]*)?(ACCESS)?[\W_]*(KEY|ACCESS[\W_]*KEY))",  # AWS_SECRET_ACCESS_KEY and variations
    r"(?i).*([\W_]*password[\W_]*).*",  # password word
]


def prepare_for_json_serialization(obj):
    """Prepare the obj recursively for JSON serialization.

    Args:
        obj: The obj.

    Returns:
        object suitable for JSON serialization.

    """
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj
    elif isinstance(obj, Mapping):
        return {
            str(prepare_for_json_serialization(k)): prepare_for_json_serialization(v)
            for k, v in obj.items()
        }
    elif isinstance(obj, Iterable):
        return [prepare_for_json_serialization(v) for v in obj]
    elif inspect.isclass(obj):
        return str(obj.__name__)
    else:
        logger.warn(f"Using __str__ for {obj} as it cannot be encoded as JSON.")
        return str(obj)


class SpiderSettingsLogging:
    @classmethod
    def from_crawler(cls, crawler):
        ext = cls()
        crawler.signals.connect(ext.spider_closed, signal=signals.spider_closed)
        return ext

    def spider_closed(self, spider):
        settings = spider.settings
        if not settings.getbool("SETTINGS_LOGGING_ENABLED"):
            return

        regex = settings.get("SETTINGS_LOGGING_REGEX")
        if regex is not None:
            settings = {k: v for k, v in settings.items() if re.search(regex, k)}
        if spider.settings.getbool("MASKED_SENSITIVE_SETTINGS_ENABLED", True):
            regex_list = spider.settings.getlist(
                "MASKED_SENSITIVE_SETTINGS_REGEX_LIST", DEFAULT_REGEXES
            )
            for reg in regex_list:
                updated_settings = {
                    k: "**********" if v else v
                    for k, v in settings.items()
                    if re.match(reg, k)
                }
                settings = {**settings, **updated_settings}

        self.output_settings(settings, spider)

    def output_settings(self, settings: dict, spider: scrapy.Spider):
        # this can be overwritten in a subclass if you want to send this data elsewhere
        indent = spider.settings.get("SETTINGS_LOGGING_INDENT")
        # convert settings obj into something representable in JSON
        json_serializable_settings = prepare_for_json_serialization(settings)
        logger.info(json.dumps(json_serializable_settings, indent=indent))



from src.scrapy_settings_log import prepare_for_json_serialization


def test_prepare_for_json_serialization_with_set():
    """Test processing a set value."""
    obj = {1, 2, 3}
    result = prepare_for_json_serialization(obj)
    assert result == [1, 2, 3]


def test_prepare_for_json_serialization_with_class_object():
    """Test processing a class object."""

    class DummyClass:
        pass

    result = prepare_for_json_serialization(DummyClass)
    assert result == "DummyClass"


def test_prepare_for_json_serialization_dict_with_class_as_key():
    """Test processing a dictionary with a class object as a key."""

    class DummyClass:
        pass

    result = prepare_for_json_serialization({DummyClass: "foo"})
    assert result == {"DummyClass": "foo"}


def test_prepare_for_json_serialization_with_nested_objects():
    """Test processing nested objects."""
    obj = {
        "foo": {
            "bar": {
                1,
                2,
                3,
            }
        }
    }
    result = prepare_for_json_serialization(obj)
    assert result == {"foo": {"bar": [1, 2, 3]}}



import logging

from scrapy import Spider
from scrapy.crawler import Crawler

from src.scrapy_settings_log import SpiderSettingsLogging

LOGGER = logging.getLogger(__name__)


def MockSpider(settings):
    crawler = Crawler(Spider, settings=settings)
    return Spider.from_crawler(crawler, name="dummy")


def test_disabled(caplog):
    spider = MockSpider({})
    logger = SpiderSettingsLogging()
    logger.spider_closed(spider)

    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert not caplog.text


def test_log_all(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    # won't check specifics here as the default settings
    # can vary with scrapy versions - presence is enough
    assert caplog.text


def test_log_filtered(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "DUMMY_INT": 4,
        "DUMMY_STR": "foo",
        "SETTINGS_LOGGING_REGEX": "DUMMY",
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '{"DUMMY_INT": 4, "DUMMY_STR": "foo"}' in caplog.text


def test_log_indented(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "DUMMY_INT": 4,
        "DUMMY_STR": "foo",
        "SETTINGS_LOGGING_REGEX": "DUMMY",
        "SETTINGS_LOGGING_INDENT": 4,
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '{\n    "DUMMY_INT": 4,\n    "DUMMY_STR": "foo"\n}' in caplog.text


def test_log_set(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "DUMMY_SET": {1, 2, 3},
        "SETTINGS_LOGGING_REGEX": "DUMMY",
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '{"DUMMY_SET": [1, 2, 3]}' in caplog.text


def test_log_custom_class(caplog):

    class CustomClass:
        pass

    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "SETTINGS_LOGGING_REGEX": "DUMMY",
        "DUMMY_CUSTOM_CLASS": {CustomClass: "/foo/bar"},
    }
    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '{"DUMMY_CUSTOM_CLASS": {"CustomClass": "/foo/bar"}}' in caplog.text


def test_log_all_should_not_return_apikey_value_by_default(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "SHUB_APIKEY": "apikey_value1",
        "shub_apikey": "apikey_value2",
        "api_key": "apikey_value3",
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '"SHUB_APIKEY": "**********"' in caplog.text
    assert '"shub_apikey": "**********"' in caplog.text
    assert '"api_key": "**********"' in caplog.text
    assert "apikey_value" not in caplog.text


def test_log_all_should_return_apikey_value_if_MASKED_SENSITIVE_SETTINGS_ENABLED_is_false(
    caplog,
):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "APIKEY": "apikey_value",
        "MASKED_SENSITIVE_SETTINGS_ENABLED": False,
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '"APIKEY": "apikey_value"' in caplog.text


def test_log_all_should_not_return_aws_secret_key_value_by_default(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "AWS_SECRET_ACCESS_KEY": "secret_value1",
        "aws_secret_access_key": "secret_value2",
        "aws_access_key": "secret_value2",
        "AWS_SECRET_KEY": "secret_value2",
        "aws_secret_key": "secret_value2",
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '"AWS_SECRET_ACCESS_KEY": "**********"' in caplog.text
    assert '"aws_secret_access_key": "**********"' in caplog.text
    assert '"aws_access_key": "**********"' in caplog.text
    assert '"AWS_SECRET_KEY": "**********"' in caplog.text
    assert '"aws_secret_key": "**********"' in caplog.text
    assert "secret_value" not in caplog.text


def test_log_all_should_not_return_password_value_by_default(caplog):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "test_password": "secret_value1",
        "PASSWORD_TEST": "secret_value2",
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert '"test_password": "**********"' in caplog.text
    assert '"PASSWORD_TEST": "**********"' in caplog.text
    assert "secret_value" not in caplog.text


def test_log_all_should_return_only_the_custom_regex_data_masked_if_MASKED_SENSITIVE_SETTINGS_REGEX_LIST_configured(
    caplog,
):
    settings = {
        "SETTINGS_LOGGING_ENABLED": True,
        "MASKED_SENSITIVE_SETTINGS_REGEX_LIST": ["apppppppikey"],
        "APIKEY": "apikey_value1",
        "apppppppikey": "some_random_value",
    }

    spider = MockSpider(settings)
    logger = SpiderSettingsLogging()
    with caplog.at_level(logging.INFO):
        logger.spider_closed(spider)

    assert "apikey_value1" in caplog.text
    assert '"apppppppikey": "**********"' in caplog.text
    assert "some_random_value" not in caplog.text
